/**
 * Copyright 2016-present, Joshua Robinson
 * All rights reserved.
 *
 * This source code is licensed under the MIT license.
 *
 */

import React, { Component, cloneElement, isValidElement } from 'react';
import ReactDOM from 'react-dom';

import adler32 from 'react-lib-adler32';
import escapeTextContentForBrowser from 'react-lib-escape-text-content-for-browser';

const __DEV__ = (process.env.NODE_ENV !== 'production');

let __MOUNTED__;

class Style extends Component {

  render() {
    if (!this.props.children) {
      return this.createStyleElement();
    }

    const styleString = this.getStyleString();
    const rootElement = this.getRootElement();

    if (!styleString && rootElement) {
      // Passthrough; no style actions
      return rootElement.props.children;
    } else if (styleString && !rootElement) {
      // Global styling with no scoping
      return this.createStyleElement(this.processCSSText(styleString));
    } else {
      // Style tree of elements
      const scopeClassName = this.getScopeClassName(styleString);

      return cloneElement(
        rootElement,
        {
          ...rootElement.props,
          className: `${rootElement.props.className ? rootElement.props.className + ' ' : ''}${scopeClassName}`
        },
        this.getNewChildrenForCloneElement(
          this.processCSSText(styleString, `.${scopeClassName}`, this.getRootSelectors(rootElement)),
          rootElement
        )
      );
    }
  }

  componentDidMount() {
    // Trip mounted flag, future style renderings are added to head instead of
    // being injected as an embedded style element
    __MOUNTED__ = true;

    if (!window._reactiveStyle) {
      const styleEl = document.createElement('style');
      styleEl.className = 'reactive-style';
      styleEl.type = 'text/css';
      styleEl.innerHTML = this._style.innerHTML;
      document.head.appendChild(styleEl);
      window._reactiveStyle = {
        el: styleEl,
        cssTextHashesAddedToHead: [adler32(this._style.innerHTML)]
      }
      this.forceUpdate();
    } else if (this._style) {
      this.addCSSTextToHead(this._style.innerHTML);
      this.forceUpdate();
    }
  }

 /**
  * Filters out the style string from this.props.children
  *
  *    > getStyleString()
  *    ".foo { color: red; }"
  *
  * @return {?string} string Style string
  */
  getStyleString = () => {
    if (this.props.children instanceof Array) {
      const styleString = this.props.children.filter((child) => (
        !isValidElement(child)
      ));

      if (styleString.length > 1) {
        throw new Error(
          'Multiple style objects as direct descedents of a ' +
          'Style component are not supported (' + styleString.length +
          ' style objects detected): \n\n' + styleString[0]
        );
      }

      return styleString[0];
    } else if (typeof this.props.children === 'string' && !isValidElement(this.props.children)) {
      return this.props.children;
    } else {
      return null;
    }
  }

  /**
  * Filters out the root element from this.props.children
  *
  *    > getRootElement()
  *    "<MyRootElement />"
  *
  * @return {?ReactDOMComponent} component Root element component
  */
  getRootElement = () => {
    if (this.props.children instanceof Array) {
      const rootElement = this.props.children.filter((child) => (
        isValidElement(child)
      ));

      if (__DEV__) {
        if(rootElement.length > 1) {
          console.log(rootElement);
          throw new Error(
            'Adjacent JSX elements must be wrapped in an enclosing tag (' +
            rootElement.length + ' root elements detected).'
          );
        }

        if (typeof rootElement[0] !== 'undefined' && this.isVoidElement(rootElement[0].type)) {
          throw new Error(
            'Self-closing void elements like ' + rootElement.type + ' must be wrapped ' +
            'in an enclosing tag. Reactive Style must be able to nest a style element ' +
            'inside of the root element and void element content models never allow' +
            'it to have contents under any circumstances.'
          );
        }
      }

      return rootElement[0];
    } else if (isValidElement(this.props.children)) {
      return this.props.children;
    } else {
      return null;
    }
  }

  /**
  * Creates an array of selectors which target the root element
  *
  *    > getRootSelectors( <div id="foo" className="bar" /> )
  *    "['#foo', '.bar']"
  *
  * @param {ReactDOMComponent} component
  * @return {!array} array Array of selectors that target the root element
  */
  getRootSelectors = (rootElement) => {
    const rootSelectors = [];

    // Handle id
    if (rootElement.props.id) {
      rootSelectors.push('#' + rootElement.props.id);
    }

    // Handle classes
    if (rootElement.props.className) {
      rootElement.props.className.trim().split(/\s+/g).forEach((className) =>
        (rootSelectors.push(className))
      );
    }

    // Handle no root selector by using type
    if (!rootSelectors.length) {
      rootSelectors.push(rootElement.type);
    }

    return rootSelectors;
  }

  /**
  * Scopes CSS statement with a given scoping class name as a union or contains selector;
  * also escapes CSS declaration bodies
  *
  *    > proccessStyleString( '.foo { color: red; } .bar { color: green; }', '_scoped-1234, ['.root', '.foo']  )
  *    ".scoped-1234.foo { color: red; } .scoped-1234 .bar { color: green; }"
  *
  * @param {string} styleString String of style rules
  * @param {string} scopedClassName Class name used to create a unique scope
  * @param {array} rootSelectors Array of selectors on the root element; ids and classNames
  * @return {!string} Scoped style rule string
  */
  processCSSText = (styleString, scopedClassName, rootSelectors) => {
    // TODO: Look into using memoizeStringOnly from fbjs/lib for escaped strings;
    // can avoid much of the computation as long as scoped doesn't come into play
    // which would be unique

    // TODO: If dev lint and provide feedback
    // if linting fails we need to error out because
    // the style string will not be parsed correctly
    return styleString
      .replace(/\s*\/\/(?![^\(]*\)).*/g, '') // Strip javascript style comments
      .replace(/\s\s+/g, ' ') // Convert multiple to single whitespace
      .split('}') // Start breaking down statements
      .map((fragment) => {
        const isDeclarationBodyPattern = /.*:.*;/g;
        const isAtRulePattern = /\s*@/g;
        const isKeyframeOffsetPattern = /\s*(([0-9][0-9]?|100)\s*%)|\s*(to|from)\s*$/g;
        // const isContent
        // Split fragment into selector and declarationBody; escape declaration body
        return fragment.split('{').map((statement) => {
          // Avoid processing whitespace
          if (!statement.trim().length) {
            return;
          }

          // Skip escaping selectors statements since that would break them;
          // note in docs that selector statements are not escaped and should
          // not be generated from user provided strings
          if (statement.match(isDeclarationBodyPattern)) {
            return escapeTextContentForBrowser(
              statement // Have to deal with special case of CSS property "content", breaks without quotes
                .replace(/lsquo|rsquo/g, '') // Prevent manipulation
                .replace(/content\s*:\s*['"](.*)['"]\s*;/, 'content: lsquo;$1rsquo;;') // "Entify" content property
                .replace(/['"]/g, '') // Remove single and double quotes
              ).replace(/lsquo;|rsquo;/g, "'") // De-"entify" content property
               .replace(/;/g, ';\n'); // Add formatting;

          } else { // Statement is a selector
            const selector = statement;

            if (scopedClassName) {
              // Prefix the scope to the selector if it is not an at-rule
              if (!selector.match(isAtRulePattern) && !selector.match(isKeyframeOffsetPattern)) {
                return this.scopeSelector(scopedClassName, selector, rootSelectors);
              } else {
                // Is at-rule or keyframe offset and should not be scoped
                return selector;
              }

            } else {
              // No scope; do nothing to the selector
              return selector;
            }
          }

        // Pretty print in dev
        }).join('{\n')
    }).join('}\n');
  }

  /**
   * Scopes a selector with a given scoping className as a union or contains selector
   *
   *    > scopeSelector( '_scoped-1827481', '.root', ['.root', '.foo']  )
   *    ".scoped-1827481.root"
   *
   * @param {string} scopedClassName Class name used to scope selectors
   * @param {string} selector Selector to scope
   * @param {array} rootSelectors Array of selectors on the root element; ids and classNames
   * @return {!string} Union or contains selector scoped with the scoping className
   */
  scopeSelector = (scopedClassName, selector, rootSelectors) => {
    let scopedSelector = '';

    // Matches comma-delimiters in multi-selectors (".fooClass, .barClass {...}" => "," );
    // ignores commas-delimiters inside of brackets and parenthesis ([attr=value], :not()..)
    const groupOfSelectorsPattern = /,(?![^\(|\[]*\)|\])/g;

    const selectors = selector.split(groupOfSelectorsPattern);

    let containsSelector; // [data-scoped="54321"] .someClass
    let unionSelector; // [data-scoped="54321"].someClass (account for root)
    for (let i = 0; i < selectors.length; i++) {
      if (rootSelectors.length && rootSelectors.some((rootSelector)=>(selector.match(rootSelector)))) {
        unionSelector = selectors[i];

        // Can't just add them together because of selector combinator complexity
        // like '.rootClassName.someClass.otherClass > *' or :not('.rootClassName'),
        // replace must be used

        // Escape valid CSS special characters that are also RegExp special characters
        const escapedRootSelectors = rootSelectors.map(rootSelector => (
          rootSelector.replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&')
        ));

        unionSelector = unionSelector.replace(new RegExp(
            '(' +                             // Start capture group
            escapedRootSelectors.join('|') +  // Match any one root selector
            ')'                               // End capture group
          )
        ,
          '$1' + scopedClassName              // Replace any one root selector match with a union
        );                                    // of the root selector and scoping class (e.g., .rootSelector._scoped-1)

        // Do both union and contains selectors because of case <div><div></div></div>
        // or <div className="foo"><div className="foo"></div></div>
        containsSelector = scopedClassName + ' ' + selectors[i];
        scopedSelector += unionSelector + ', ' + containsSelector;
      } else {
        containsSelector = scopedClassName + ' ' + selectors[i];
        scopedSelector += containsSelector;
      }
    }

    return scopedSelector;
  }

  /**
   * Creates a className used as a CSS scope by generating a checksum from a styleString
   *
   *    > scoped( 'footer { color: red; }' )
   *    "_scoped-182938591"
   *
   * @param {string} String of style rules
   * @return {!string} A scoping class name
   */
  getScopeClassName = (styleString) => ('_scoped-' + adler32(styleString));

  /**
   * Checks if a tag type is a self-closing void element
   *
   *    > isVoidElement( "img" )
   *    "true"
   *
   * @param {*} string Element type to check
   * @return {!bool} bool True or false
   */
  isVoidElement = (type) => (
    [
      'area',
      'base',
      'br',
      'col',
      'command',
      'embed',
      'hr',
      'img',
      'input',
      'keygen',
      'link',
      'meta',
      'param',
      'source',
      'track',
      'wbr'
    ].some((voidType) => (type === voidType))
  )

  /**
   * Add CSS text to the style element in the head of document unless it has
   * already been added.
   *
   *    > addCSSTextToHead( ".foo { color: red; }" )
   *
   * @param {string} string CSS text to add to head
   */
  addCSSTextToHead = (cssText) => {
    if (!cssText.length) {
      return;
    } else {
      const cssTextHash = adler32(cssText);

      if (!window._reactiveStyle.cssTextHashesAddedToHead.some((hash) => (hash === cssTextHash))) {
        window._reactiveStyle.el.innerHTML += cssText
        window._reactiveStyle.cssTextHashesAddedToHead.push(cssTextHash);
      }
    }
  }

  /**
  * Creates the style element used for server side rendering
  *    > createStyleElement( ".foo._scoped-1 { color: red; }" )
  *
  *
  * @param {string} string CSS string
  * @return {ReactDOMComponent} component
  */
  createStyleElement = (cssText) => {
      return <style type="text/css" key="-0" ref={(c) => (this._style = c)}
        dangerouslySetInnerHTML={{
          __html: cssText || ''
      }}/>
  }

  /**
  * Returns new children for a root element being cloned. If mounted the CSS text
  * is added to the style element in head, otherwise we are doing server side rendering
  * and to avoid flash of unstyled content (FOUC) a style element is added to children
  * to avoid FOUC on first render.
  *
  *    > getNewChildrenForCloneElement( ".foo._scoped-1 { color: red; }" )
  *     "<NewChildren />"
  *
  * @param {string} string CSS string
  * @return {ReactDOMComponent} component
  */
  getNewChildrenForCloneElement = (cssText, rootElement) => {
    if (__MOUNTED__) {
      return rootElement.props.children;
    } else {
      return [this.createStyleElement(cssText)].concat(rootElement.props.children)
    }
  }

  /**
  * Syntactic sugar for functional usage of Reactive Style
  *
  *    > Style.it( ".foo { color: red; }", <div /> )
  *     "<div class="_scoped-1">
  *        <style type="text/css">
  *          .foo._scoped-1 { color: red; }
  *        </style>
  *      </div>"
  *
  * @param {string} string CSS string
  * @param {ReactDOMComponent} component
  */
  static it = (cssText, rootElement) => {
    return (
      <Style>
        {cssText}
        {rootElement}
      </Style>
    )
  }
}

export default Style;